"""
Jenkins Core module.

This module loads a file like "https://updates.jenkins.io/stable-2.190/latestCore.txt"
to determine the latest "stable" or "current" Jenkins version.
It then provides methods through a class to retrieve the latest jenkins package
on "https://pkg.jenkins.io/".
"""

import logging
import os
import urllib.request

logger = logging.getLogger(__name__)


DEFAULT_JENKINS_CORE_URL = "https://updates.jenkins.io"
DEFAULT_JENKINS_REPO = "https://pkg.jenkins.io"
DEFAULT_JENKINS_RELEASE = "stable"


class JenkinsCore(object):
    """Represents Jenkins core manager."""

    def __init__(
        self,
        jenkins_core_url=DEFAULT_JENKINS_CORE_URL,
        jenkins_repo=DEFAULT_JENKINS_REPO,
        jenkins_release=DEFAULT_JENKINS_RELEASE,
    ):
        """Instantiate the JenkinsCore class.

        Load the latest version of the `jenkins_release` available.

        :param str jenkins_core_url: Url where the latest Jenkins core version is set.
        :param str jenkins_repo: Url where the Jenkins binary packages are stored.
        :param str jenkins_release: Flavor of Jenkins. Values are "stable" or "current".
        """
        # TODO: Use python-apt to use the real repository data (and manage correct checksum)
        # Right now this manage only the official jenkins repository layout, which is non-standard
        self.jenkins_repo = jenkins_repo
        if jenkins_release not in ("current", "stable"):
            err = "Wrong jenkins_release parameter. Valid values are [current|stable]"
            logger.error(err)
            raise ValueError(err)
        self.jenkins_release = jenkins_release
        self.latest_core_url = "{0}/{1}/latestCore.txt".format(jenkins_core_url, jenkins_release)
        self.core_version = self.load_latest_core_version(self.latest_core_url)

    def load_latest_core_version(self, core_url):
        """Load the latest version available.

        :param str core_url: Url of the latest Jenkins core
        """
        logger.info("Loading %s", core_url)
        with urllib.request.urlopen(core_url) as core_data:
            core_version = core_data.read().decode("utf-8")
            logger.debug("Latest jenkins core version: %s", core_version)
        return core_version

    def get_binary_package(self, dst_dir, distribution="debian", version=None, arch="amd64"):
        """Get a Jenkins binary package.

        Download a binary package of the said distribution and of the latest Jenkins
        release if not specified.

        :param str dst_dir: Target directory for the download
        :param str distribution: Distribution family to retrieve package for (debian, redhat)
        :param str version: The version you want to download. Default to the latest Core if not specified
        :param str arch: Architecture to download. Default to amd64
        """

        assert distribution in ("debian", "redhat")
        if self.jenkins_release == "stable":
            repo_dist = "{0}-stable".format(distribution)
        else:
            repo_dist = distribution

        if not version:
            target_version = self.core_version
        else:
            target_version = version
        binary_package_url = "{url}/{dist}/binary/jenkins_{version}_all.deb".format(
            url=self.jenkins_repo, dist=repo_dist, version=target_version
        )
        package_file = os.path.basename(binary_package_url)
        package_file_path = os.path.join(dst_dir, package_file)
        with urllib.request.urlopen(binary_package_url) as binary_pkg:
            if not os.path.exists(os.path.dirname(package_file_path)):
                os.makedirs(os.path.dirname(package_file_path))
            logger.info("Downloading %s to %s", binary_package_url, package_file_path)
            with open(package_file_path, "wb") as target_pkg:
                target_pkg.write(binary_pkg.read())
            logger.info("%s downloaded", package_file)
