#!/usr/bin/env python3
"""
Entrypoint for the jenkins-plugin-manager module
"""

import argparse
import logging
import sys

from jenkins_plugin_manager.plugin import UpdateCenter


logger = logging.getLogger("jenkins_plugin_manager")


def parse_args():
    """Parse the arguments provided to the CLI."""
    parser = argparse.ArgumentParser(prog="jenkins-plugin-manager")
    parser.add_argument("--log-level", action="store", help="Logging level", default="WARNING")
    parser.add_argument("--log-file", action="store", help="Log file")
    parser.add_argument("--log-file-level", action="store", help="File logging level", default="INFO")
    subparsers = parser.add_subparsers(help="sub-command help")

    # Sub command "plugin"
    parser_plugin = subparsers.add_parser("plugin", help="Manage Jenkins plugins")
    parser_plugin.add_argument("--from-file", action="store", help="Load a list of plugin from a file")
    subparser_plugin = parser_plugin.add_subparsers(help="plugin sub-command help")

    # Sub command "plugin list-urls"
    parser_plugin_list = subparser_plugin.add_parser("list-urls", help="List Jenkins plugins url")
    parser_plugin_list.add_argument("args", nargs=argparse.REMAINDER, help="List of plugin")
    parser_plugin_list.set_defaults(func=plugin_list_urls)

    # Sub command "plugin download"
    parser_plugin_download = subparser_plugin.add_parser("download", help="Download Jenkins plugins")
    parser_plugin_download.add_argument(
        "--dst", action="store", required=True, help="Directory for the downloaded plugins"
    )
    parser_plugin_download.add_argument("args", nargs=argparse.REMAINDER, help="List of plugin")
    parser_plugin_download.set_defaults(func=plugin_download)

    return parser.parse_args()


def configure_logging(args):
    """Configure the logging.

    :param :py:class:argparse.ArgumentParser args: Arguments provided
    """
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    logger.setLevel(logging.DEBUG)
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        raise ValueError("Invalid log-level level: %s" % args.log_level)
    console = logging.StreamHandler()
    console.setLevel(log_level)
    console.setFormatter(formatter)
    logger.addHandler(console)
    if args.log_file:
        log_file_level = getattr(logging, args.log_file_level.upper(), None)
        if not isinstance(log_file_level, int):
            raise ValueError("Invalid --log-file-level level: %s" % args.log_file_level)
        logfile_handler = logging.FileHandler(args.log_file)
        logfile_handler.setLevel(log_file_level)
        logfile_handler.setFormatter(formatter)
        logger.addHandler(logfile_handler)


def plugin_load(args):
    """Return the list of plugin provided by the options.

    :param :py:class:argparse.ArgumentParser args: Arguments provided
    """
    plugins = []
    if args.from_file:
        try:
            with open(args.from_file, "r") as plugin_file:
                plugins.extend(plugin_file.read().splitlines())
        except FileNotFoundError:
            logger.error("File %s: Not found", args.from_file)
            sys.exit(1)
    if args.args:
        plugins.extend(args.args)
    if not plugins:
        logger.error("No plugin provided.")
        sys.exit(1)
    return plugins


def plugin_download(args):
    """Download the given plugins.

    :param :py:class:argparse.ArgumentParser args: Arguments provided
    """
    plugins = plugin_load(args)
    uc = UpdateCenter()
    for plugin in uc.get_plugins(set(plugins)):
        print(uc.download_plugin(plugin, args.dst))


def plugin_list_urls(args):
    """Sub-command "plugin list-urls".

    Print the url of each plugin that need is specified.

    :param :py:class:argparse.ArgumentParser args: Arguments provided
    """
    plugins = plugin_load(args)
    uc = UpdateCenter()
    for plugin in uc.get_plugins(set(plugins)):
        print(uc.get_plugin_data(plugin)["url"])


def main():
    """Main entrypoint for jenkins-plugin-manager CLI tool."""
    args = parse_args()
    configure_logging(args)
    args.func(args)


if __name__ == "__main__":
    main()
